import numpy as np
#import scipy.io.wavfile
import matplotlib.pyplot as plt
import scipy.io
import os


import SER_Datasets_Libs.SER_DB as SER_DB
from FormantsLib.FormantsExtract import Extract_files_formant_features as Extract_formants
import FormantsLib.FormatsHDFread as HDFread
import PhonemeSER as PhonemeSER

                              
def split_train_n_test(features, labels, scheme='single-random', test_size=0.2, folds_n=5, test_speakers=5, features_2=None, labels_2=None): 
    '''
    Parameters:

    features: array-like, dtype=int16, shape = [total_clips, max_frames, formants*4]. Array of clip frames formant features;

    labels: array-like, dtype=int16, shape = [total_clips, 11]. Each clip has 11 labels including speaker_id, emotions, scenario, frame_count, trimmed_length etc.

    scheme : string. Choose one of these {'L1SO', 'LMSO',  'k-folds', 'single-random', 'cross-corpus'}, default: 'single-random',;

    test_speakers : only for scheme = 'LMSO', Leave Many Speakers Out;

    folds_n : only for scheme = 'k-folds';

    test_size : only for scheme = 'single-random';

    features_2, labels_2: Only for scheme = 'cross-corpus';

    returns: features_train, labels_train, features_test, labels_test

    '''

    features_train = []
    labels_train = []
    features_test = []
    labels_test = []
    if(scheme=='L1SO'):
        
        print("Validation scheme:", 'L1SO')
        unique_speakers = np.unique(labels[:, HDFread.Ix.speaker_id])

        for s_id in range(len(unique_speakers)):
            select_train = np.where(labels[:, HDFread.Ix.speaker_id]!=unique_speakers[s_id])
            select_test = np.where(labels[:, HDFread.Ix.speaker_id]==unique_speakers[s_id])

            features_train.append(features[select_train] )
            labels_train.append(labels[select_train] )
            features_test.append(features[select_test] )
            labels_test.append(labels[select_test])
            #print ("Test Speaker:", unique_speakers[s_id], "Train:", features_train[s_id].shape[0], "Test:", features_test[s_id].shape[0])
    elif(scheme=='LMSO'):
        
        print("Validation scheme:", 'LMSO', test_speakers)
        unique_speakers = np.unique(labels[:, HDFread.Ix.speaker_id])
        total_speakers = len(unique_speakers)
        folds = int(round(total_speakers/test_speakers))

        for s_id in range(folds):
            stx = test_speakers*s_id
            edx = stx + test_speakers
            if edx > total_speakers: edx = total_speakers
            this_fold_speakers = unique_speakers[stx:edx]

            conditions = labels[:, HDFread.Ix.speaker_id] == this_fold_speakers[0]
            for ts in this_fold_speakers:
                conditions1 =  (labels[:, HDFread.Ix.speaker_id] == ts)
                conditions = conditions | conditions1
            
            select_train = np.where(conditions != True)
            select_test = np.where(conditions == True)

            features_train.append(features[select_train] )
            labels_train.append(labels[select_train] )
            features_test.append(features[select_test] )
            labels_test.append(labels[select_test])
            print ("Test speakers:", this_fold_speakers, "Train:", features_train[s_id].shape[0], "Test:", features_test[s_id].shape[0])
            
    elif(scheme=='k-folds'):
        print("Validation scheme:", 'k-folds', folds_n)
        from sklearn.model_selection import ShuffleSplit
        rs = ShuffleSplit(n_splits=folds_n, test_size=1/folds_n, random_state=0)
        for select_train, select_test in rs.split(features):
            
            features_train.append(features[select_train] )
            labels_train.append(labels[select_train] )
            features_test.append(features[select_test] )
            labels_test.append(labels[select_test])
            #print ("Fold:", "Train:", features[select_train].shape[0], "Test:", features[select_test].shape[0])

    elif(scheme=='single-random'):
        print("Validation scheme:", 'single-random', test_size)
        from sklearn.model_selection import train_test_split
        X_train, X_test, Y_train, Y_test = train_test_split(features, labels, test_size = test_size, random_state = 100)
    
        features_train.append(X_train)
        labels_train.append(Y_train)
        features_test.append(X_test)
        labels_test.append(Y_test)

    elif(scheme=='cross-corpus'):
        print("Validation scheme:", 'cross-corpus')
        if (features_2.all()!=None and labels_2.all()!=None):
            print("Whole DB_1 is (X_train, Y_train), DB_2 is (X_test, Y_test)")

            features_train.append(features)
            labels_train.append(labels)
            features_test.append(features_2)
            labels_test.append(labels_2)
        else:
            raise Exception ("Cross-corpus missing arguments: 'features_2' and 'labels_2'")
    else:
        raise Exception ("Invalid argument. Select any of there validation schemes: 'L1SO', 'LMSO',  'k-folds', 'single-random', 'cross-corpus'")


    return features_train, labels_train, features_test, labels_test


def save_results_csv(results_csv, db_names, classifier_name, uar, war, scheme, test_size, folds_n, K_SD, features_count, Nm_inst, Nm_diff, g_dist, window_length, window_step, emphasize_ratio, classes_n, speakers_n, samples, confusion):
    import csv

    import os
    if not os.path.exists('results'):
        os.makedirs('results')
    if(os.path.isfile("results\\" + results_csv)==False):
        with open("results\\" + results_csv, 'w') as csvFile:
            csvFile.write("db_names, classifier, UAR, WAR, scheme, test_size, folds_n, K_SD, features_count, Nm_inst, Nm_diff, g_dist, window_length, window_step, emphasize_ratio, classes_n, speakers_n, samples, confusion\n")

    with open("results\\" + results_csv, 'a') as csvFile:
        writer = csv.writer(csvFile, delimiter=',', lineterminator = '\n')
        #writer.writerow(["Emotion", "Combination", "Occurrences"])
        this_row = [db_names, classifier_name, str(uar), str(war), scheme, str(test_size), str(folds_n), str(K_SD), 
                str(features_count), str(Nm_inst), str(Nm_diff),
                str(g_dist), str(window_length*1000), str(window_step*1000), str(emphasize_ratio), 
                str(classes_n), str(speakers_n), str(samples), str(confusion)]


        writer.writerow(this_row)

def make_formants_filename(db_name, window_length, window_step, emphasize_ratio, norm):
    if not os.path.exists('data'):
        os.makedirs('data')
    return 'data\\Formants_' + db_name + '_'  + str(int(window_length*1000)) + '_' + str(int(window_step*1000)) + '_' + str(int(emphasize_ratio*1000)) + '_' + str(norm) + '.hdf'

def make_model_filename(DB_names, val_scheme, test_size, folds_n, test_speakers, K_SD, Nm_inst, Nm_diff, g_dist, window_length, window_step, emphasize_ratio, norm):
    if not os.path.exists('data'):
        os.makedirs('data')
    return "data\\model_" + DB_names + "_" + str(val_scheme) + str(test_size)+ "_"+ str(folds_n) + "_"+ str(test_speakers) +"_" + str(K_SD) + "_" + "I".join(str(x) for x in Nm_inst) + "D".join(str(x) for x in Nm_diff) + str(g_dist) + str(window_length) + str(window_step) + str(emphasize_ratio)+ '_' + str(norm)+ ".pkl"

def new_type_of_labels(db_name_path, deselect_labels, labels):
    array_of_clips = SER_DB.create_DB_file_objects(db_name_path['DB'], db_name_path['path'], deselect=deselect_labels)
    if(len(array_of_clips)!=labels.shape[0]): raise Exception("Can't edit labels. Array size mismatch. Recreate HDF to create labels data with new (and perhaps changed) conditions.")

    import csv
    file_i = []
    file_C6 = []
    with open('C:\\DB\\IEMOCAP_relabel\\C4.csv') as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            file_i.append(row[list(row.keys())[0]]) #error while using 'i' as key
            file_C6.append(row['C4'])

    for i in range (0, labels.shape[0]):
        labels[i, HDFread.Ix.emotion] = int(file_C6[file_i.index(os.path.basename(array_of_clips[i].filepath))])

    print(labels[0, HDFread.Ix.speaker_id], labels[0, HDFread.Ix.emotion], labels[0, HDFread.Ix.scenario])
    print(array_of_clips[0].speaker_id, ord(array_of_clips[0].emotion_cat), array_of_clips[0].scenario)
    return labels
    


def run_train_and_test(db_names_paths, results_csv, val_scheme='single-random', test_size=0.2, folds_n=5, test_speakers=2, Nm_inst=[32, 64], Nm_diff=[32, 64], K_SD=0, g_dist=8, window_length=0.025, window_step=0.01, emphasize_ratio=0.65, norm=0, deselect_labels=None, db_names_paths_2=None):
    '''
    Run training and testing functions with the set parameters

    Parameters
    ----------

    `db_names_paths`: Dict list of names and paths of DBs. Shape: [{'DB':'<db_name>', 'path': '<dir_path>'},], Choose db_name from these: {"EmoDB", "RAVDESS", "IEMOCAP", "ShemoDB", "DEMoS"}
    with respective directory paths of the DB wav files. Importing of each DB is programed in 'Dataset_Lib/SER_DB.py'. The organization format of annotated wav files is expected to be as is downloaded from their original sources. For val_scheme='cross-corpus', all these DBs are used as training sets and `db_names_paths_2` must be passed as testing DB list.

    `results_csv`: string, optional (default=None), CSV file path to which final results will be appended.

    `val_scheme`: {'L1SO', 'LMSO', 'single-random', 'k-folds', 'cross-corpus'}, optional (default=single-random). Select one of these validation schemes: 'L1SO' : Leave One Speaker Out, 'LMSO': Leave multiple speakers out (default: 5 speakers, require parameter 'test_speakers'), 'single-random': Single random split (requires param 'test_size'), 'k-folds': K-folds cross-validation (require param: 'folds_n').

    `test_size`: float, optional (default=0.2). Fraction of total clips to use as testing set for 'single-random' validation scheme. Ignored for other schemes.

    `folds_n`: unsigned int, optional (default=5). Number of folds for 'k-folds' validation scheme. Ignored for other schemes.

    `test_speakers`: unsigned int, optional (default=5). Number of test speakers for 'LMSO' validation scheme. Ignored for other schemes.

    `Nm_inst`: array-like, dtype=int16, shape = [n_models], optional (default=[32, 64]). Cluster numbers for instantaneous (~25ms) phoneme clustering model. Set between 8 to 300 for each cluster model.
    
    `Nm_diff`:  array-like, dtype=int16, shape = [n_models], optional (default=[32, 64]). Cluster numbers for differential phoneme (~25ms * g_dist) clustering models. Set between 8 to 300 for each cluster model.

    `K_SD`: float, optional (default=0.0). Feature selection parameters. Set between -1 to 1. It sets the limit of standard deviation below the mean for selecting features within this threshold. Lower value selects more features.

    `g_dist`: unsigned int, optional (default=8). Number for adjacent frames for measuring the change in formant features to calculate differential phoneme features.
    
    `emphasize_ratio`: float, optional (default=0.7). Amplitude increasing factor for pre-emphasis of higher frequencies (high frequencies * emphasize_ratio = balanced amplitude as low frequencies).

    `norm`: int, optional, (default=0), Enable or disable normalization of Mel-filters.

    `deselect_labels`: list of chars, optional (default=None). Example: deselect_labels=['F', 'B'] to deselect 'F'and 'B'.

    `db_names_paths_2`: Dict list of of test DBs, shape: [{'DB':'<db_name>', 'path': '<dir_path>'},], optional (default=None). Same as 'db_names_paths', but only required for 'cross-corpus' validation scheme.

    Returns
    -------

    `return 0` : on succussful execution.
    '''
    
    #list of HDF storage file paths in which formant characteristics are stored
    features_HDF_files = []
    features_test_HDF_files = []

    
    DB_names = ".".join([db_names_paths[x]['DB'][0:4] for x in range(0, len(db_names_paths))])

    

    # Create the model filename using all the parameters.
    models_save_file = make_model_filename(DB_names, val_scheme, test_size, folds_n, test_speakers, K_SD, Nm_inst, Nm_diff, g_dist, window_length, window_step, emphasize_ratio, norm)

    
    
    #check if features are already extracted
    for db_name_path in db_names_paths:
        #HDF storage file path in which formant characteristics are stored
        features_HDF_file = make_formants_filename(db_name_path['DB'], window_length, window_step, emphasize_ratio, norm)
        print(features_HDF_file)
        if (os.path.isfile(features_HDF_file)==False) or (int(os.path.getsize(features_HDF_file))<8000):
            array_of_clips = SER_DB.create_DB_file_objects(db_name_path['DB'], db_name_path['path'], deselect=deselect_labels)
            #Extract and save formant features of clips in array to an HDF file along with labels (labels are included in file_objects)
            processed_n = Extract_formants(array_of_clips, features_HDF_file, window_length, window_step, emphasize_ratio, norm, f0_min=30, f0_max=4000, max_frames=800, formants=3)
            if(processed_n==0):
                raise Exception("No files to process. Make sure DB directory path and filenames are in the correct format.")
        
        features_HDF_files.append(features_HDF_file)

    #Read formant features and labels from the HDF file
    features, labels, u_speakers, u_classes  = HDFread.import_mutiple_HDFs(features_HDF_files, deselect_labels=deselect_labels)

    
    #labels = new_type_of_labels(db_names_paths[0], deselect_labels, labels)
        
    HDFread.print_database_stats(labels)

    features2, labels2 = None, None
    #Only for cross-corpus validation scheme
    if(val_scheme=='cross-corpus'):
        if(len(db_names_paths_2)>0):
            for db_name_path in db_names_paths_2:
                features_HDF_file = make_formants_filename(db_name_path['DB'], window_length, window_step, emphasize_ratio, norm)

                if (os.path.isfile(features_HDF_file)==False) or (int(os.path.getsize(features_HDF_file))<8000):
                    array_of_clips = SER_DB.create_DB_file_objects(db_name_path['DB'], db_name_path['path'], deselect=deselect_labels)
                    processed_n = Extract_formants(array_of_clips, features_HDF_file, window_length, window_step, emphasize_ratio, f0_min=30, f0_max=4000, max_frames=800, formants=3)
                    if(processed_n==0):
                        raise Exception("No files to process. Make sure DB directory path and filenames are in the correct format.")
                features_test_HDF_files.append(features_HDF_file)
            features2, labels2, u_speakers2, u_classes2  = HDFread.import_mutiple_HDFs(features_test_HDF_files, deselect_labels=deselect_labels)
            print("Train set labels:", [chr(x) for x in u_classes])
            #HDFread.print_database_stats(labels1)
            print("Test set labels:", [chr(x) for x in u_classes])
            DB_names += "_"
            DB_names += ".".join([db_names_paths_2[x]['DB'][0:4] for x in range(0, len(db_names_paths_2))])
        else:
            raise Exception ("Missing argument 'db_names_paths_2'")




    folds_results = []
    features_counts = []
    
    Xtr, Ytr, Xts, Yts = split_train_n_test(features, labels, val_scheme, test_size, folds_n, test_speakers, features_2=features2, labels_2=labels2)

    for fold in range(0, len(Xtr)):
        print ("Fold:", fold, "Train:", Xtr[fold].shape[0], "Test:", Xts[fold].shape[0])

        print("Training")
        if((val_scheme!='cross-corpus') or ((val_scheme=='cross-corpus') and (os.path.isfile(models_save_file)==False))):
            PhonemeSER.Train_model(models_save_file, Xtr[fold], Ytr[fold][:, HDFread.Ix.emotion], Ytr[fold][:, HDFread.Ix.frame_count], Nm_inst, Nm_diff, K_SD, g_dist)
        else:
            print("Skipping training because model file already exists.")
        
        print("Testing")
        print("Using model file:", models_save_file, "\t Test samples:", Xts[fold].shape[0])
        classifiers_results, ft_count = PhonemeSER.Test_model(models_save_file, Xts[fold], Yts[fold][:, HDFread.Ix.emotion], Yts[fold][:, HDFread.Ix.frame_count])
        
    
        folds_results.append(classifiers_results)
        features_counts.append(ft_count)
    
    features_count = int(np.mean(features_counts)) #mean of features counts of each fold

    #classifiers =  [folds_results[x]['classifier'] for x in range(0, len(folds_results[0]))]
    print("Folds:", len(folds_results), "\rClassifiers:", len(folds_results[0]), "\tMean features:", features_count)
    
    for c in range(len(folds_results[0])):
        sum_conf = np.zeros_like(folds_results[0][c]['confusion'])
        for fold in range(len(folds_results)):
            sum_conf += folds_results[fold][c]['confusion']
    
        
            
        samples = np.sum(sum_conf)
        print(folds_results[0][c]['classifier'], "\tUAR:", folds_results[0][c]['UAR'], " WAR:", folds_results[0][c]['WAR'], "\tTested samples:", samples)
        
        if(results_csv!=None):
            save_results_csv(results_csv, DB_names, folds_results[0][c]['classifier'], folds_results[0][c]['UAR'], folds_results[0][c]['WAR'], val_scheme, test_size, folds_n, K_SD, features_count, Nm_inst, Nm_diff, g_dist, window_length, window_step, emphasize_ratio, len(u_classes), len(u_speakers), samples, sum_conf)

    print("Finished 'run_train_and_test' for DB", DB_names)

    return 0





def main():



    #Few parameters:
    pt = [16, 32, 64, 128]  #total phoneme clusters, add more than one integer to this list to create multiple models
    K_SD = -0.5             #float, -1 to +1, feature selection parameters, less K_SD selects more featues
    norm = 0            #Normalization of mel-filter banks, doesn't work at 1 (9-12-2020)

    
    # List of DB names and directory where wav files are stored:
    db_names_paths = [{'DB': "EmoDB", 'path' : "E:\\LABWORK\\Datasets\\EMO-DB\\wav\\"},
    {'DB': "RAVDESS", 'path' : "E:\\LABWORK\\Datasets\\RAVDESS_COPY2\\"},
    {'DB': "IEMOCAP", 'path' : "E:\\LABWORK\\Datasets\\IEMOCAP_noVideo\\"},
    {'DB': "ShemoDB", 'path' : "E:\\LABWORK\\Datasets\\shemo\\"},
    {'DB': "DEMoS", 'path' : "E:\\LABWORK\\Datasets\\wav_DEMoS\\"},
    {'DB': "MSPIMPROV", 'path' : "E:\\LABWORK\\Datasets\\MSP-IMPROV\\"}]
    # Can add more than one DBs to the list

    '''
    Currently supports these DBs
    {'DB': "EmoDB", 'path' : "E:\\LABWORK\\Datasets\\EMO-DB\\wav\\"},
    {'DB': "RAVDESS", 'path' : "E:\\LABWORK\\Datasets\\RAVDESS_COPY2\\"},
    {'DB': "IEMOCAP", 'path' : "E:\\LABWORK\\Datasets\\IEMOCAP_noVideo\\"},
    {'DB': "ShemoDB", 'path' : "E:\\LABWORK\\Datasets\\shemo\\"},
    {'DB': "DEMoS", 'path' : "E:\\LABWORK\\Datasets\\wav_DEMoS\\"},
    {'DB': "MSPIMPROV", 'path' : "E:\\LABWORK\\Datasets\\MSP-IMPROV\\"}
    '''
    
    deselect_labels=['D','F','U','E','R', 'C', 'G', 'B', 'X']
    run_train_and_test(db_names_paths, "results.csv", val_scheme='single-random', test_size=0.1, folds_n=5, test_speakers=2, Nm_inst=pt, Nm_diff=pt, K_SD=K_SD, g_dist=6, window_length=0.025, window_step=0.010, emphasize_ratio=0.65, norm=norm, deselect_labels=deselect_labels)
    
    exit()
    
    
    # For cross-corpus validation:
    # Training sets:
    db_names_paths = [{'DB': "MSPIMPROV", 'path' : "C:\\DB\\MSP-IMPROV\\"},]

    #Testing sets
    db_names_paths_2 =  [{'DB': "IEMOCAP", 'path' : "C:\\DB\\IEMOCAP_noVideo\\"},]

    
    deselect_labels=['D','F','U','E','R', 'C', 'G', 'B', 'X']

    run_train_and_test(db_names_paths, "results.csv", val_scheme='cross-corpus', Nm_inst=pt, Nm_diff=pt, K_SD=K_SD, g_dist=6, window_length=0.025, window_step=0.01, emphasize_ratio=0.65, norm=norm, deselect_labels=deselect_labels, db_names_paths_2=db_names_paths_2)
    

    


    print("Done")
    exit()
    



if __name__ == '__main__':
    main()





"""
-----
Author: Abdul Rehman
License:  The MIT License (MIT)
Copyright (c) 2020, Tabahi Abdul Rehman
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
"""

from os.path import basename
import numpy as np
from scipy import signal as signallib
#from numba import jit #install numba to speed up the execution
from pydub import AudioSegment

#@jit(nopython=True) 
def get_lowest_positions(array_y, n_positions):
    order = array_y.argsort()
    ranks = order.argsort() #ascending
    top_indexes = np.zeros((n_positions,), dtype=np.int16)
    #print(array_y)
    i = int(0)

    while(i < n_positions):
        itemindices = np.where(ranks==i)
        for itemindex in itemindices:
            if(itemindex.size):
                #print(i, array_y[itemindex], itemindex)
                top_indexes[i] = itemindex[0]
            else:   #for when positions are more than array size
                itemindices2 = np.where(ranks==(array_y.size -1-i+ array_y.size ))
                for itemindex2 in itemindices2:
                    #print(i, array_y[itemindex2], itemindex2)
                    top_indexes[i] = itemindex2[0]
            i += 1
    #print(array_y[top_indexes])
    return top_indexes


#@jit(nopython=True) 
def get_top_positions(array_y, n_positions):
    order = array_y.argsort()
    ranks = order.argsort() #ascending
    top_indexes = np.zeros((n_positions,), dtype=np.int16)
    #print(array_y)
    i = int(n_positions - 1)

    while(i >= 0):
        itemindices = np.where(ranks==(len(array_y)-1-i))
        for itemindex in itemindices:
            if(itemindex.size):
                #print(i, array_y[itemindex], itemindex)
                top_indexes[i] = itemindex[0]
            else:   #for when positions are more than array size
                itemindices2 = np.where(ranks==len(array_y)-1-i+len(array_y) )
                for itemindex2 in itemindices2:
                    #print(i, array_y[itemindex2], itemindex2)
                    top_indexes[i] = itemindex2[0]
            i -= 1

    return top_indexes
    

def frame_segmentation(signal, sample_rate, window_length=0.040, window_step=0.020):

    #Framing
    frame_length, frame_step = window_length * sample_rate, window_step * sample_rate  # Convert from seconds to samples
    signal_length = len(signal)
    frame_length = int(round(frame_length))
    frame_step = int(round(frame_step))
    num_frames = int(np.ceil(float(np.abs(signal_length - frame_length)) / frame_step))  # Make sure that we have at least 1 frame

    if(num_frames < 1):
        raise Exception("Clip length is too short. It should be atleast " + str(window_length*2)+ " frames")

    pad_signal_length = num_frames * frame_step + frame_length
    z = np.zeros((pad_signal_length - signal_length))
    pad_signal = np.append(signal, z) # Pad Signal to make sure that all frames have equal number of samples without truncating any samples from the original signal

    indices = np.tile(np.arange(0, frame_length), (num_frames, 1)) + np.tile(np.arange(0, num_frames * frame_step, frame_step), (frame_length, 1)).T
    frames = pad_signal[indices.astype(np.int32, copy=False)]

    #Hamming Window
    frames *= np.hamming(frame_length)
    #frames *= 0.54 - 0.46 * numpy.cos((2 * numpy.pi * n) / (frame_length - 1))  # Explicit Implementation **
    #print (frames.shape)
    return frames, signal_length


def get_filter_banks(frames, sample_rate, f0_min=60, f0_max=4000, num_filt=128, norm=0):
    '''
    Fourier-Transform and Power Spectrum

    return filter_banks, hz_points

    filter_banks: array-like, shape = [n_frames, num_filt]

    hz_points: array-like, shape = [num_filt], center frequency of mel-filters

    This code is from https://haythamfayek.com/2016/04/21/speech-processing-for-machine-learning.html
    Courtesy of Haytham Fayek
    '''

    NFFT = num_filt*32      #FFT bins (equally spaced - Unlike mel filter)
    mag_frames = np.absolute(np.fft.rfft(frames, NFFT))  # Magnitude of the FFT
    pow_frames = ((1.0 / NFFT) * ((mag_frames) ** 2))  # Power Spectrum

    #Filter Banks
    nfilt = num_filt
    low_freq_mel = (2595 * np.log10(1 + (f0_min) / 700))
    high_freq_mel = (2595 * np.log10(1 + (f0_max) / 700))  # Convert Hz to Mel
    mel_points = np.linspace(low_freq_mel, high_freq_mel, nfilt + 2)  # Equally spaced in Mel scale
    hz_points = (700 * (10**(mel_points / 2595) - 1))  # Convert Mel to Hz
    bin = np.floor((NFFT + 1) * hz_points / sample_rate)

    n_overlap = int(np.floor(NFFT / 2 + 1))
    fbank = np.zeros((nfilt, n_overlap))
    
    for m in range(1, nfilt + 1):
        f_m_minus = int(bin[m - 1])   # left
        f_m = int(bin[m])             # center
        f_m_plus = int(bin[m + 1])    # right
        
        for k in range(f_m_minus, f_m):
            fbank[m - 1, k] = (k - bin[m - 1]) / (bin[m] - bin[m - 1])
        for k in range(f_m, f_m_plus):
            fbank[m - 1, k] = (bin[m + 1] - k) / (bin[m + 1] - bin[m])
    filter_banks = np.dot(pow_frames, fbank.T)
    filter_banks = np.where(filter_banks == 0, np.finfo(float).eps, filter_banks)  # Numerical Stability
    #filter_banks = 20 * np.log10(filter_banks)  # dB
    if(norm):
        filter_banks -= (np.mean(filter_banks)) #normalize

    return filter_banks, hz_points




freq, power, width, dissonance = 0,1,2,3



def Extract_formant_descriptors(fft_x, fft_y, formants=2, f_min=30, f_max=4000):
    '''
    returns 12D-array, shape = ((formants*4,), dtype=np.uint64)
    '''
    
    len_of_x = len(fft_x)
    len_of_y = len(fft_y)
    
    #for 4 features
    returno = np.zeros((formants*4,), dtype=np.uint64)

    if(len_of_x!=len_of_y) or (len_of_x<=3):
        #print("Empty Frame")
        return returno

    peak_indices = signallib.argrelextrema(fft_y, np.greater, mode='wrap')
    valley_indices = signallib.argrelextrema(fft_y, np.less, mode='wrap')
    peak_indices = peak_indices[0]
    peak_fft_x, peak_fft_y = fft_x[peak_indices], fft_y[peak_indices]
    valley_fft_x, valley_fft_y = fft_x[valley_indices], fft_y[valley_indices]

    
    len_of_peaks = len(peak_indices)
    if(len_of_peaks < 1) or (len(valley_indices) < 1):
        #print("Silence")
        return returno


    ground_level = 0
    if (len(valley_fft_y) > 1):
        ground_level = np.max(valley_fft_y)  #range(valleys_y)/2
    if(ground_level<10):
        #Silence
        return returno
    
    #add extra valleys at start and end
    if(peak_fft_x[0] < valley_fft_x[0]):
        valley_fft_x = np.append([f_min/2], valley_fft_x)
        valley_fft_y = np.append([ground_level/8], valley_fft_y)
    if(peak_fft_x[-1] > valley_fft_x[-1]):
        valley_fft_x = np.append(valley_fft_x, [f_max+f_min])
        valley_fft_y = np.append(valley_fft_y, [ground_level/8])

    top_peaks_n = formants*2
    #make sure fft has enought points
    
    if(len(peak_fft_y)<(formants+1)):
        return returno
    if(len(peak_fft_y)<(top_peaks_n-1)):
        top_peaks_n = len(peak_fft_y) - 1

    tp_indexes = get_top_positions(peak_fft_y, top_peaks_n) #descending
    dissonance_peak = np.zeros(top_peaks_n)
    biggest_peak_y = peak_fft_y[tp_indexes[0]]
    
    formants_detected = 0

    #calc width and dissonance
    for i in range(0, top_peaks_n):
        
        if(dissonance_peak[i]==0) and (peak_fft_y[tp_indexes[i]] > (biggest_peak_y/16))  and (peak_fft_x[tp_indexes[i]] >= f_min) and (peak_fft_x[tp_indexes[i]] <= f_max) and (formants_detected < formants):
            next_valley = np.min(np.where(valley_fft_x > peak_fft_x[tp_indexes[i]]))
            next_valley_x = valley_fft_x[next_valley]
            next_valley_y = valley_fft_y[next_valley]

            this_peak_gnd_thresh = peak_fft_y[tp_indexes[i]]/4

            
            while(next_valley_y > this_peak_gnd_thresh) and (len(np.where(valley_fft_x > next_valley_x)[0])>0):
                valley_next_peak_ind = np.where(peak_fft_x > next_valley_x)
                if(len(valley_next_peak_ind[0])>0):
                    valley_next_peak = np.min(valley_next_peak_ind)
                    if(peak_fft_y[tp_indexes[i]] > peak_fft_y[valley_next_peak]):
                        next_valley = np.min(np.where(valley_fft_x > next_valley_x))
                        next_valley_x = valley_fft_x[next_valley]
                        next_valley_y = valley_fft_y[next_valley]
                    else:
                        break
                else:
                    break
                
                
                        
            prev_valley = np.max(np.where(valley_fft_x < peak_fft_x[tp_indexes[i]]))
            prev_valley_x = valley_fft_x[prev_valley]
            prev_valley_y = valley_fft_y[prev_valley]

            while(prev_valley_y > this_peak_gnd_thresh) and (len(np.where(valley_fft_x < prev_valley_x)[0])>0):
                valleys_prev_peak_ind = np.where(peak_fft_x < prev_valley)
                if(len(valleys_prev_peak_ind[0])>0):
                    valley_prev_peak = np.max(valleys_prev_peak_ind)
                    if(peak_fft_y[tp_indexes[i]] > peak_fft_y[valley_prev_peak]):
                        prev_valley = np.max(np.where(valley_fft_x < prev_valley_x))
                        prev_valley_x = valley_fft_x[prev_valley]
                        prev_valley_y = valley_fft_y[prev_valley]
                    else:
                        break
                else:
                    break


            dissonance_peak[i] = 1
            this_dissonane = 0
            for k in range(0, top_peaks_n):
                if(peak_fft_x[tp_indexes[k]] < next_valley_x) and (peak_fft_x[tp_indexes[k]] > prev_valley_x) and k!=i:
                    dissonance_peak[k] = 1
                    if(np.abs(peak_fft_x[tp_indexes[k]] - peak_fft_x[tp_indexes[i]]) > (peak_fft_x[tp_indexes[i]]/50)):
                        this_dissonane += peak_fft_y[tp_indexes[k]]
                    else:
                        peak_fft_x[tp_indexes[i]] = (peak_fft_x[tp_indexes[i]]+peak_fft_x[tp_indexes[k]])/2
                        peak_fft_y[tp_indexes[i]] = (peak_fft_y[tp_indexes[i]]+peak_fft_y[tp_indexes[k]])/2
            

            this_dissonane = this_dissonane/peak_fft_y[tp_indexes[i]]
            this_width = np.log(next_valley_x)-np.log(prev_valley_x)
            

            returno[freq + (formants_detected*4)] = peak_fft_x[tp_indexes[i]]
            returno[power + (formants_detected*4)] = peak_fft_y[tp_indexes[i]]
            returno[width + (formants_detected*4)] = this_width*10
            returno[dissonance + (formants_detected*4)] = this_dissonane*10
            
            
            formants_detected += 1

             
    #plt.figure(1)
    #plt.plot(fft_x, fft_y)
    #plt.plot(peak_fft_x, peak_fft_y, marker='o', linestyle='dashed', color='green', label="Splits")
    #plt.plot(valley_fft_x, valley_fft_y, marker='o', linestyle='dashed', color='red', label="Splits")
    #plt.show()

    
    return returno

    


def normalize_volume(wav_file_path):
    sound = AudioSegment.from_file(wav_file_path, "wav")
    change_in_dBFS = -20.0 - sound.dBFS
    normalized_sound =  sound.apply_gain(change_in_dBFS)
    temp_file_path = wav_file_path + ".x"
    normalized_sound.export(temp_file_path, format="wav")
    return temp_file_path

def Extract_wav_file_formants(wav_file_path, window_length=0.025, window_step=0.010, emphasize_ratio=0.7, norm=0, f0_min=30, f0_max=4000, max_frames=400, formants=3, formant_decay=0.5):
    '''
    Parameters
    ----------

    `wav_file_path`: string, Path of the input wav audio file;

    `window_length`: float, optional (default=0.025). Frame window size in seconds;

    `window_step`: float, optional (default=0.010). Frame window step size in seconds;

    `emphasize_ratio`: float, optional (default=0.7). Amplitude increasing factor for pre-emphasis of higher frequencies (high frequencies * emphasize_ratio = balanced amplitude as low frequencies);

    `norm`: int, optional, (default=0), Enable or disable normalization of Mel-filters;

    `f0_min`: int, optional, (default=30), Hertz;

    `f0_max`: int, optional, (default=4000), Hertz;
    
    `max_frames`: int, optional (default=400). Cut off size for the number of frames per clip. It is used to standardize the size of clips during processing.
    
    `formants`: int, optional (default=3). Number of formants to extract;

    `formant_decay`: float, optional (default=0.5). Decay constant to exponentially decrease feature values by their formant amplitude ranks;

    Returns
    -------
    returns `frames_features, frame_count, signal_length, trimmed_length`

    `frames_features`: array-like, `np.array((max_frames, num_of_features*formants), dtype=np.uint16)`

    `frame_count`: int, number of filled frames (out of max_frames);

    `signal_length`: float, signal length in seconds;

    `trimmed_length`: float, trimmed length in seconds, silence at the begining and end of the input signal is trimmed before processing;
    '''

    
    from wavio import read as wavio_read
    wav_data = wavio_read(wav_file_path)
    raw_signal = wav_data.data
    sample_rate = wav_data.rate

    #emphasize_ratio = 0.70
    signal_to_plot = np.append(raw_signal[0], raw_signal[1:] - emphasize_ratio * raw_signal[:-1])
    #signal_to_plot = raw_signal
    
    num_filt = 256
    frames, signal_length = frame_segmentation(signal_to_plot, sample_rate, window_length=window_length, window_step=window_step)
    frames_filter_banks, hz_points = get_filter_banks(frames, sample_rate, f0_min=f0_min, f0_max=f0_max, num_filt=num_filt, norm=norm)
    
    #x-axis points for triangular mel filter used
    #hz_bins_min = hz_points[0:num_filt] #discarding last 2 points
    hz_bins_mid = hz_points[1:num_filt+1] #discarding 1st and last point
    #hz_bins_max = hz_points[2:num_filt+2] #discarding first 2 points

    
    num_of_frames = frames_filter_banks.shape[0]

    #min_peaks_count = 2
    
    neighboring_frames = 2  #number of neighboring frames to compares
    if(num_of_frames < ((neighboring_frames*2)+1)):
        raise Exception("Not enough frames to compare harmonics. Need at least" + str(neighboring_frames*2)+ " frames. Frame count:", str(num_of_frames))

    #formants = 2
    num_of_features = 4 #freq, power, width, dissonance
    formants_data = np.zeros((num_of_frames, num_of_features*formants), dtype=np.uint64)
    
    for frame_index in range(0, num_of_frames): #except first and last 5 frames
      
        # Find peaks(max).
        peak_indexes = signallib.argrelextrema(frames_filter_banks[frame_index], np.greater, mode='wrap')
        peak_indexes = peak_indexes[0]
        peak_fft_x, peak_fft_y = hz_bins_mid[peak_indexes], frames_filter_banks[frame_index][peak_indexes]

        formants_data[frame_index] = Extract_formant_descriptors(peak_fft_x, peak_fft_y, formants, f0_min, f0_max)

    
    
    #mean(power of 1st formant)/40
    power_ground = int(np.mean(formants_data[:,power][np.where(formants_data[:,power] > 0)])/1000)
    if(power_ground<1):
        power_ground = 1
    
    

    #trim silent ends
    first_frame, last_frame = 0, 0
    for i in range(0,num_of_frames):
        first_frame = i
        if(formants_data[i, power]>power_ground):
            break

    for i in range(0, num_of_frames):
        last_frame = num_of_frames - i - 1
        if(formants_data[last_frame, power]>power_ground):
            break

    #print(power_ground, num_of_frames, last_frame - first_frame)
    trimmed_length = ((last_frame - first_frame)/num_of_frames)*signal_length

    

    #convert to db
    for fr in range(0, num_of_frames):
        for i in range(0, formants):
            formant_decay_rate = formant_decay**(i)
            
            if(formants_data[fr, power + (i*num_of_features)] < 1):
                formants_data[fr, power + (i*num_of_features)] = 0
            else:
                formants_data[fr, power + (i*num_of_features)] = np.log10(formants_data[fr, power + (i*num_of_features)]) * 100 * formant_decay_rate
            
            if(formants_data[fr, freq + (i*num_of_features)] < f0_min):
                formants_data[fr, freq + (i*num_of_features)] = 0
            else:
                formants_data[fr, freq + (i*num_of_features)] = np.log(formants_data[fr, freq + (i*num_of_features)]) * 200 * formant_decay_rate
            
            formants_data[fr, width + (i*num_of_features)] = formants_data[fr, width + (i*num_of_features)] * 5 * formant_decay_rate
            formants_data[fr, dissonance + (i*num_of_features)] = formants_data[fr, dissonance + (i*num_of_features)] * 10 * formant_decay_rate
        #print(formants_data[fr])
        #exit()
    returno = np.zeros((max_frames, num_of_features*formants), dtype=np.uint16)
    frame_count = 0
    for i in range(0, max_frames):
        old_frame_i = first_frame+i
        returno[i] = formants_data[old_frame_i]
        frame_count = i
        if(i >= (last_frame - first_frame - 1)):
            break
        elif(i >= (max_frames-1)):
            print("Warning! Frame size overflow, Size:", (last_frame - first_frame), "Limit:", max_frames)
            break

    #print(frame_count, signal_length/sample_rate, trimmed_length/sample_rate)
    return returno, frame_count, signal_length/sample_rate, trimmed_length/sample_rate




def Extract_files_formant_features(array_of_clips, features_save_file, window_length=0.025, window_step=0.010, emphasize_ratio=0.7, norm=0, f0_min=30, f0_max=4000, max_frames=400, formants=3,):
    '''
    Parameters
    ----------
    `array_of_clips`: list of Clip_file_Class objects from 'SER_DB.py';

    `features_save_file`: string, Path for HDF file where extracted features will be stored;

    `window_length`: float, optional (default=0.025). Frame window size in seconds;

    `window_step`: float, optional (default=0.010). Frame window step size in seconds;

    `emphasize_ratio`: float, optional (default=0.7). Amplitude increasing factor for pre-emphasis of higher frequencies (high frequencies * emphasize_ratio = balanced amplitude as low frequencies);

    `norm`: int, optional, (default=0), Enable or disable normalization of Mel-filters;

    `f0_min`: int, optional, (default=30), Hertz;

    `f0_max`: int, optional, (default=4000), Hertz;
    
    `max_frames`: int, optional (default=400). Cut off size for the number of frames per clip. It is used to standardize the size of clips during processing.
    
    `formants`: int, optional (default=3). Number of formants to extract;

    returns processed_clips
    ----------------------

    processed_clips: int, number of successfully processing clips;
    '''

    import os
    if(os.path.isfile(features_save_file)):
        print("Removing HDF")
        os.remove(features_save_file)


    total_clips = len(array_of_clips)
    processed_clips = 0
    
    import h5py
    with h5py.File(features_save_file, 'w') as hf:
        dset_label = hf.create_dataset('labels', (total_clips, 11),  dtype='u2')
        dset_features = hf.create_dataset('features', (total_clips, max_frames, formants*4), dtype='u2')
        
        print("Clip", "i", "of", "Total", "SpeakerID", "Scenario", "Sex", "Emotion")
        for index, clip in enumerate(array_of_clips):
            try:
                print("Clip ", index+1, "of", total_clips, clip.speaker_id, clip.scenario, clip.sex, clip.emotion_cat)
                array_frames_by_features = np.zeros((max_frames, formants*4), dtype=np.uint16)
                #print(clip.filepath)
                temp_file_path = normalize_volume(clip.filepath)

                array_frames_by_features, frame_count, signal_length, trimmed_length = Extract_wav_file_formants(temp_file_path, window_length, window_step, emphasize_ratio, norm, f0_min, f0_max, max_frames, formants)
                clipfile_size = int(os.path.getsize(clip.filepath)/1000)
                
                os.remove(temp_file_path)
                dset_features[index] = array_frames_by_features
                dset_label[index] = [clip.speaker_id, clip.scenario, ord(clip.sex), ord(clip.emotion_cat), int(clip.intensity_cat), int(clip.statement), int(clip.repetition), int(frame_count), int(signal_length*1000), int(trimmed_length*1000), clipfile_size]
                processed_clips += 1
            except Exception as e:
                print (e)
            
        print("Read features of", total_clips, "clips")
    
    print("Closing HDF")
    return processed_clips






"""
-----
Author: Abdul Rehman
License:  The MIT License (MIT)
Copyright (c) 2020, Tabahi Abdul Rehman
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.


Dependencies: sklearn (v2.1), numba (v0.45.1), pickle, matplotlib
"""

import numpy as np
#from numba import jit #install numba to speed up the execution
from timeit import default_timer as timer

from sklearn.neural_network import MLPClassifier
from sklearn.svm import SVC
#from sklearn.svm import LinearSVC
#from sklearn.svm import NuSVC
#from sklearn.gaussian_process import GaussianProcessClassifier
#from sklearn.gaussian_process.kernels import RBF
#from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier
#from sklearn.ensemble import AdaBoostClassifier
#from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis
#from sklearn.linear_model import LogisticRegression
from sklearn.neighbors import KNeighborsClassifier
from sklearn.ensemble import BaggingClassifier
from sklearn.tree import DecisionTreeClassifier


'''
Confusion matrix, UAR, WAR calcultions
'''

def create_conf_matrix(expected, predicted, classes):
        n_classes = len(classes)
        m = np.zeros((n_classes, n_classes), dtype=np.uint16)

        for i in range(len(expected)):
            for exp in range(n_classes):
                for pred in range(n_classes):
                    if(expected[i]==classes[exp]) and (predicted[i]==classes[pred]):
                        m[exp][pred] += 1
        return m



def get_UAR_WAR(confusion_matrix):
    class_recalls = np.array([])
    class_weights = np.array([])
    for row in range(0, len(confusion_matrix)):
        if(np.nansum(confusion_matrix[row]) > 0):
            class_recalls = np.append(class_recalls, confusion_matrix[row, row] / np.nansum(confusion_matrix[row]))
        else:
            class_recalls = np.append(class_recalls, confusion_matrix[row, row] * 0)
        class_weights = np.append(class_weights, np.nansum(confusion_matrix[row]) / np.nansum(confusion_matrix))
    uar = np.mean(class_recalls)
    war = np.sum(class_recalls*class_weights)

    return round(uar*100, 2), round(war*100, 2)


#@jit(nopython=True) # Set "nopython" mode for best performance, equivalent to @njit
def balance_data(X, Y):

    u_labels = np.unique(Y)
    print(u_labels)
    if(X.shape[0] != Y.shape[0]): raise Exception("X and Y length mismatch")
    #print("Balancing data for labels", [(chr(int(lbl)) if lbl > 50 else lbl) for lbl in u_labels])
    cat_sizes = np.zeros(len(u_labels))
    

    for u1 in range(0, len(u_labels)):   cat_sizes[u1] = len(np.where(Y == u_labels[u1])[0])
    print(cat_sizes)

    
    max_size = int(np.max(cat_sizes))
    
    for u in range(0, len(u_labels)):
        cat_Y_sel = np.where(Y == u_labels[u])[0]
        #np.random.shuffle(cat_Y_sel)    #shuffle after label selection
        u_size = len(cat_Y_sel)
        cat_Y = Y[cat_Y_sel]
        cat_X = X[cat_Y_sel]
        it = 0

        Y_u_new = np.empty((max_size - u_size,))
        X_u_new = np.empty((max_size - u_size, X.shape[1]))
        
        Y_new = np.empty((Y.shape[0] + max_size - u_size, ))
        X_new = np.empty((X.shape[0] + max_size - u_size, X.shape[1]))

        while(u_size < max_size):
            #new_Y = np.append(new_Y, cat_Y[it])
            Y_u_new[max_size - u_size - 1] = cat_Y[it]
            #concatenate workaround
            X_u_new[max_size - u_size - 1, :] = cat_X[it]
            #new_X = np.concatenate([new_X, np.array([cat_X[it]])])
            
            u_size += 1
            it += 1
            if (it >= cat_sizes[u]): it = 0
        
        #print(max_size, X.shape, cat_X.shape, X_u_new.shape)

        Y_new[0:X.shape[0]] = Y
        Y_new[X.shape[0]:] = Y_u_new
        Y = Y_new
        Y_new = None
        
        X_new[0:X.shape[0], :] = X
        X_new[X.shape[0]:, :] = X_u_new
        X = X_new
        X_new = None

    for u1 in range(0, len(u_labels)):   cat_sizes[u1] = len(np.where(Y == u_labels[u1])[0])
    print(cat_sizes)
    
    return X, Y


'''
Classifier functions and objects
'''

def Test(classifier_model, X_features):

    from sklearn import preprocessing
    X_features = preprocessing.scale(X_features, with_mean=False, with_std=True, axis=0)
    y_pred = classifier_model.predict(X_features)
    #test_acc = accuracy_score(Y_labels, y_pred) # same as WAR
    #results_array.append(test_acc)
    
    
    return y_pred
    '''
    
            'RF02': RandomForestClassifier(max_depth=10, n_estimators=100, max_features=10),
            'RF03': RandomForestClassifier(max_depth=30, n_estimators=100, max_features=7),
            'RF04': RandomForestClassifier(max_depth=30, n_estimators=100, max_features=10),
            'RF05': RandomForestClassifier(max_depth=30, n_estimators=100, max_features=10),
    '''

class ClassifierObject(object):

    classifiers = {
        
            'SVC1': SVC(decision_function_shape='ovr', kernel='rbf', C=10, gamma='scale', probability=False, class_weight='balanced',),
            'SVC2': SVC(decision_function_shape='ovr', kernel='rbf', C=1, gamma='scale', probability=False, class_weight='balanced'),
            'SVC3': SVC(decision_function_shape='ovr', kernel='rbf', C=0.1, gamma='scale', probability=False, class_weight='balanced'),
            'SVC4': SVC(decision_function_shape='ovr', kernel='rbf', C=0.01, gamma='scale', probability=False, class_weight='balanced'),
            'SVC5': SVC(kernel='poly', degree=3, C=1,  gamma='scale', random_state=0, probability=False),
            'RF01': RandomForestClassifier(max_depth=10, n_estimators=100, max_features=7),
            'RF03': RandomForestClassifier(max_depth=30, n_estimators=10, max_features=7),
            'RF04': RandomForestClassifier(max_depth=30, n_estimators=100, max_features=10),
            'KNN1': KNeighborsClassifier(n_neighbors=20, weights='uniform', leaf_size=30, p=2, metric='minkowski'),
            'MLP1': MLPClassifier(alpha=1, max_iter=2000, validation_fraction=0.3, hidden_layer_sizes=(400,100,50,), )
            
        }


    def Train(self, X_features, Y_labels):

        from sklearn import preprocessing
        X_features = preprocessing.scale(X_features, with_mean=False, with_std=True, axis=0)

        X_features, Y_labels = balance_data(X_features, np.array(Y_labels, float))

        #from sklearn.model_selection import train_test_split
        #X_train, X_test, Y_train, Y_test = train_test_split(X_features, Y_labels, test_size = validation, random_state = 100)
        #from sklearn.metrics import accuracy_score

        #classifiers['SVR2'].fit(X_train, Y_train)
        for index, (name, classifier) in enumerate(self.classifiers.items()):
            self.classifiers[name].fit(X_features, Y_labels)
            #scores = cross_val_score(classifier, X_train, Y_train_label_ovr, cv=5)
            #y_pred = classifier.predict(X_test)
            #test_acc = accuracy_score(Y_test, y_pred) # same as WAR
            

        return self.classifiers




'''
Feature selection
'''

#@jit(nopython=True) 
def get_top_positions(array_y, n_positions):
    order = array_y.argsort()
    ranks = order.argsort() #ascending
    top_indexes = np.zeros((n_positions,), dtype=np.int16)
    #print(array_y)
    i = int(n_positions - 1)

    while(i >= 0):
        itemindices = np.where(ranks==(len(array_y)-1-i))
        for itemindex in itemindices:
            if(itemindex.size):
                #print(i, array_y[itemindex], itemindex)
                top_indexes[i] = itemindex[0]
            else:   #for when positions are more than array size
                itemindices2 = np.where(ranks==len(array_y)-1-i+len(array_y) )
                for itemindex2 in itemindices2:
                    #print(i, array_y[itemindex2], itemindex2)
                    top_indexes[i] = itemindex2[0]
            i -= 1

    return top_indexes
    


def select_features(train_features, Y_labels_train, K_SD=-0.5, mode=0):
    features_n = train_features.shape[1]
    label_classes = np.unique(Y_labels_train)

    if (mode==1):
        #select same number of features optimized for each emotion.
        select_n = int(features_n*np.abs(1-K_SD)/len(label_classes))
        selected_indices = np.zeros((select_n*len(label_classes), ), dtype=np.uint16)

        for e in range(0, len(label_classes)):
            select_emo = np.where(Y_labels_train==label_classes[e])

            features_std_within_emo= np.std(train_features[select_emo], axis=0)
            selected_indices[select_n*e:select_n*(e+1)] = get_top_positions(features_std_within_emo, select_n)
        return selected_indices
    
    elif (mode==2):

        selected_indices = np.zeros(([]), dtype=np.uint16)

        from sklearn.model_selection import ShuffleSplit
        splits_n = 4
        rs = ShuffleSplit(n_splits=splits_n, test_size=None, random_state=0)
        
        feature_means = np.zeros((len(label_classes), splits_n, features_n,), dtype=np.float)
        feature_means_per_emo = np.zeros((len(label_classes), features_n,), dtype=np.float)
        feature_std_per_emo = np.zeros((len(label_classes), features_n,), dtype=np.float)
        #parts_emo_mean_std = np.zeros((splits_n, features_n,), dtype=np.float)
        tsp = 0
        for big_part, s_part in rs.split(train_features):
            for e in range(0, len(label_classes)):
                # select part --> select emotion --> features mean
                train_features_part = train_features[s_part]
                Y_labels_train_part = Y_labels_train[s_part]
                select_emo = np.where(Y_labels_train_part==label_classes[e])
                feature_means[e, tsp, :] = np.mean(train_features_part[select_emo], axis=0)
            # part --> features std across emotions
            tsp += 1
        
        for e in range(0, len(label_classes)):
            feature_means_per_emo[e] = np.mean(feature_means[e], axis=0)
            feature_std_per_emo[e] = np.std(feature_means[e], axis=0)
        
        for e in range(0, len(label_classes)):

            features_mean_values = np.mean(feature_means_per_emo, axis=0)
            emo_mean_diff = np.abs(features_mean_values - feature_means_per_emo[e])

            K = K_SD
            thresh_mean = np.mean(emo_mean_diff) + (np.std(feature_std_per_emo[e])*(K))
            thresh_std = np.mean(feature_std_per_emo[e]) + (np.std(feature_std_per_emo[e])*(1 - K))

            conditions = (emo_mean_diff > thresh_mean) & (feature_std_per_emo[e] < thresh_std)

            #selected_indices = np.where(conditions)[0]
            selected_indices = np.append(selected_indices, np.where(conditions)[0])
            #print(np.where(conditions)[0].size)

        return np.unique(selected_indices)

    else:
        #select features with high SD across emotions
        feature_means = np.zeros((len(label_classes), features_n, ), dtype=np.float)
        #features_std = np.zeros((len(label_classes), features_n, ), dtype=np.float)
        for e in range(0, len(label_classes)):
            select_emo = np.where(Y_labels_train==label_classes[e])
            feature_means[e, :] = np.mean(train_features[select_emo], axis=0)
            #features_std[e, :] = np.std(train_features[select_emo], axis=0)

        emo_mean_std = np.std(feature_means, axis=0)
        
        #selected_indices = get_top_positions(emo_std_mean, selected_features_n)
        thresh = np.mean(emo_mean_std) + (np.std(emo_mean_std) * K_SD)
        selected_indices = np.where(emo_mean_std > thresh)[0]
    
        if (len(selected_indices) < 1):
            return range(0,features_n)
            
        return selected_indices      




'''
Phoneme clustering
'''

def make_clusters(cluster_n1, frames_formants, clustering_method=0, print_quality_scores=0, save_plot=False):

    
    #duration = timer() - start
    #print('Time: %0.1fs' % duration)
    start = timer()

    max_frames = frames_formants.shape[1]
    X_Clips_N = frames_formants.shape[0]
    
    max_selected_clips = int(X_Clips_N*0.2)
    if(max_selected_clips < 50): max_selected_clips = int(X_Clips_N*0.8)
    random_select = []
    import random
    while (len(random_select) < (max_selected_clips)):
        n = random.randint(0,X_Clips_N-1)
        if(n not in random_select):
            random_select.append(n)
    
    max_selected_clips = len(random_select)

    combined_frames = np.zeros((max_selected_clips*max_frames, frames_formants.shape[2]), dtype=np.uint16)

    #print("Clustering data shape:", combined_frames.shape)

    for i in range(0, max_selected_clips):
        for k in range(0, max_frames):
            combined_frames[(i*max_frames) + k, :] = frames_formants[random_select[i], k,:]
    
    clustModel = None

     
    #from sklearn.cluster import Birch
    #clustModel = Birch(n_clusters=None, threshold=0.5, branching_factor=500)
    #clustModel.fit(combined_frames)

    metric='euclidean'
    if(clustering_method==0):
        from sklearn.cluster import MiniBatchKMeans
        clustModel = MiniBatchKMeans(cluster_n1, init='k-means++', batch_size=2000, max_iter=500, n_init=5, compute_labels=save_plot)
        clustModel.fit(combined_frames)
        clustModel.fit(combined_frames)

    elif(clustering_method==1):
        from sklearn.cluster import MiniBatchKMeans
        clustModel = MiniBatchKMeans(cluster_n1, init='k-means++', batch_size=1000, max_iter=500, n_init=5, compute_labels=save_plot)
        clustModel.fit(combined_frames)
        clustModel.fit(combined_frames)
    
    elif(clustering_method==2):
        from sklearn.cluster import MiniBatchKMeans
        clustModel = MiniBatchKMeans(cluster_n1, init='k-means++', batch_size=2000, max_iter=500, n_init=10, compute_labels=save_plot)
        clustModel.fit(combined_frames)
        clustModel.fit(combined_frames)
    
    elif(clustering_method==3):
        from sklearn.cluster import MiniBatchKMeans
        clustModel = MiniBatchKMeans(cluster_n1, init='random', batch_size=1000, max_iter=500, n_init=10, compute_labels=save_plot)
        clustModel.fit(combined_frames)
        clustModel.fit(combined_frames)
        
    elif(clustering_method==4):
        from sklearn.cluster import KMeans
        clustModel = KMeans(cluster_n1, init='k-means++', max_iter=500)
        clustModel.fit(combined_frames)
        
    elif(clustering_method==5):
        from sklearn.cluster import AgglomerativeClustering
        clustModel = AgglomerativeClustering(n_clusters=cluster_n1, linkage="ward", affinity=metric) # only "euclidean" with ward is accepted
        clustModel.fit(combined_frames)

    elif(clustering_method==6):
        from sklearn.cluster import AgglomerativeClustering
        clustModel = AgglomerativeClustering(n_clusters=cluster_n1, linkage="average", affinity=metric) # “euclidean”, “l1”, “l2”, “manhattan”, “cosine”, or “precomputed”
        clustModel.fit(combined_frames)

    elif(clustering_method==7):
        metric='l1'
        from sklearn.cluster import AgglomerativeClustering
        clustModel = AgglomerativeClustering(n_clusters=cluster_n1, linkage="average", affinity=metric) # “euclidean”, “l1”, “l2”, “manhattan”, “cosine”, or “precomputed”
        clustModel.fit(combined_frames)
    
    elif(clustering_method==8):
        metric='l2'
        from sklearn.cluster import AgglomerativeClustering
        clustModel = AgglomerativeClustering(n_clusters=cluster_n1, linkage="average", affinity=metric) # “euclidean”, “l1”, “l2”, “manhattan”, “cosine”, or “precomputed”
        clustModel.fit(combined_frames)
    
    elif(clustering_method==9):
        metric='manhattan'
        from sklearn.cluster import AgglomerativeClustering
        clustModel = AgglomerativeClustering(n_clusters=cluster_n1, linkage="average", affinity=metric) # “euclidean”, “l1”, “l2”, “manhattan”, “cosine”, or “precomputed”
        clustModel.fit(combined_frames)



    

    

    if(print_quality_scores):
        print("Calculating clustering quality scores")
        duration = timer() - start
        
        unique_phonemes = len(np.unique(clustModel.labels_))

        if(unique_phonemes < 2):
            print("Error: Only 1 cluster is created. Please try different clustering parameters.")

        else:
            from sklearn import metrics
            silh_s = metrics.silhouette_score(combined_frames, clustModel.labels_, metric=metric)
            ch_s = metrics.calinski_harabasz_score(combined_frames, clustModel.labels_)
            db_s = metrics.davies_bouldin_score(combined_frames, clustModel.labels_)
            #See details at:
            #https://scikit-learn.org/stable/modules/clustering.html#silhouette-coefficient

            # higher Silhouette Coefficient score relates to a model with better defined clusters
            # Calinski-Harabasz The score is higher when clusters are dense and well separated, which relates to a standard concept of a cluster.
            # Davies-Bouldin Index: Zero is the lowest possible score. Values closer to zero indicate a better partition.
            print('Clustering, Time, Unique Cluster Lables, Silhouette, Calinski-Harabasz, Davies-Bouldin')
            print('%d, %0.1fs,  %d, %0.3f, %d, %0.3f' % (clustering_method, duration, unique_phonemes, silh_s, ch_s, db_s))

            

    if(save_plot):
        print("Creating cluster scatter plot")

        import matplotlib.pyplot as plt

        fig = plt.figure(figsize=(12, 10))
        fig.subplots_adjust(left=0.08, right=0.98, bottom=0.09, top=0.99)
        ax1 = fig.add_subplot(2, 2, 1)
        ax2 = fig.add_subplot(2, 2, 2)
        ax3 = fig.add_subplot(2, 2, 3)
        ax4 = fig.add_subplot(2, 2, 4)

        #ax1.set_xlim([600, 1700])
        ax2.set_ylim([64000, 66000])
        #ax3.set_xlim([65000, 70000])
        ax4.set_ylim([0, 2000])
        
        ax1.scatter(combined_frames[:, 0], combined_frames[:, 1], c=clustModel.labels_, cmap=plt.cm.nipy_spectral, alpha=0.25)
        ax1.set_xlabel('f\u2080', fontsize=12, fontstyle='italic', fontfamily='serif')
        ax1.set_ylabel('p\u2080', fontsize=12, fontstyle='italic', fontfamily='serif')

        #ax2.scatter(combined_frames[:, 0], combined_frames[:, 4], c=clustModel.labels_, cmap=plt.cm.nipy_spectral, alpha=0.25)
        #ax2.set_xlabel('f\u2080', fontsize=12, fontstyle='italic', fontfamily='serif')
        #ax2.set_ylabel('f\u2081', fontsize=12, fontstyle='italic', fontfamily='serif')

        ax2.scatter(combined_frames[:, 0], combined_frames[:, 2], c=clustModel.labels_, cmap=plt.cm.nipy_spectral, alpha=0.25)
        ax2.set_xlabel('f\u2080', fontsize=12, fontstyle='italic', fontfamily='serif')
        ax2.set_ylabel('Diff-f\u2080', fontsize=12, fontstyle='italic', fontfamily='serif')

        #ax3.scatter(combined_frames[:, 0], combined_frames[:, 2], c=clustModel.labels_, cmap=plt.cm.nipy_spectral, alpha=0.25)
        #ax3.set_xlabel('f\u2080', fontsize=12, fontstyle='italic', fontfamily='serif')
        #ax3.set_ylabel('w\u2080', fontsize=12, fontstyle='italic', fontfamily='serif')

        ax3.scatter(combined_frames[:, 1], combined_frames[:, 3], c=clustModel.labels_, cmap=plt.cm.nipy_spectral, alpha=0.25)
        ax3.set_xlabel('p\u2080', fontsize=12, fontstyle='italic', fontfamily='serif')
        ax3.set_ylabel('Diff-p\u2080', fontsize=12, fontstyle='italic', fontfamily='serif')

        #ax4.scatter(combined_frames[:, 0], combined_frames[:, 3], c=clustModel.labels_, cmap=plt.cm.nipy_spectral, alpha=0.25)
        #ax4.set_xlabel('f\u2080', fontsize=12, fontstyle='italic', fontfamily='serif')
        #ax4.set_ylabel('d\u2080', fontsize=12, fontstyle='italic', fontfamily='serif')

        
        ax4.scatter(combined_frames[:, 0], combined_frames[:, 2], c=clustModel.labels_, cmap=plt.cm.nipy_spectral, alpha=0.25)
        ax4.set_xlabel('f\u2080', fontsize=12, fontstyle='italic', fontfamily='serif')
        ax4.set_ylabel('Diff-f\u2080', fontsize=12, fontstyle='italic', fontfamily='serif')


        plt.savefig("clusters_" + str(cluster_n1) + "_" + str(clustering_method) + ".png", dpi=300)
        #plt.show()
        
        
    
    return clustModel







#@jit(nopython=True)
def formants_slope(frames_formants, frames_len, distance):
    max_clips = frames_formants.shape[0]
    max_frames = frames_formants.shape[1]
    #max_formants = 2
    slope_ftr = np.zeros((max_clips, max_frames, 4,))
    half_dist = int(distance/2)
    
    for i in range(0, max_clips):
        for frm in range(0, frames_len[i] - distance):
            slope_ftr[i, frm, 0] = np.mean(frames_formants[i, frm:frm+distance, 0])
            slope_ftr[i, frm, 1] = np.mean(frames_formants[i, frm:frm+distance, 1])
            #slope_ftr[i, frm, 2] = np.mean(frames_formants[i, frm:frm+distance, 4])
            
            slope_ftr[i, frm, 2] = (np.mean(frames_formants[i, frm:frm+half_dist, 0]) - np.mean(frames_formants[i, frm+half_dist:frm+distance, 0]))
            slope_ftr[i, frm, 3] = np.mean(frames_formants[i, frm:frm+half_dist, 1]) - np.mean(frames_formants[i, frm+half_dist:frm+distance, 1])
            #slope_ftr[i, frm, 5] = (np.abs(np.mean(frames_formants[i, frm:frm+half_dist, 4]) - np.mean(frames_formants[i, frm+half_dist:frm+distance, 4])))

    return slope_ftr


#@jit(nopython=True)
def count_phonemes_in_clip(phonemes_array, frames_len, phoneme_types):
    phoneme_count = np.zeros((phoneme_types,))
    
    for frm in range(0, frames_len):
        phoneme_count[phonemes_array[frm]] += 1000
    
    phoneme_count = phoneme_count/(frames_len+1)
    return phoneme_count


def count_phonemes_in_set(clustModel, frames_formants, frames_len,  phoneme_types):
    max_clips = frames_formants.shape[0]
    clips_phoneme_count = np.zeros((max_clips, phoneme_types,), dtype=np.float)

    for i in range(0, max_clips):
        phonemes_array = clustModel.predict(frames_formants[i])
        clips_phoneme_count[i] = count_phonemes_in_clip(phonemes_array, frames_len[i], phoneme_types)

    return clips_phoneme_count






def Train_model(models_save_file, X_formants_train, Y_labels_train, X_frame_lens_train, inst_phoneme_types=[16,32,64], diff_phoneme_types=[16,32,64], K_SD=0, g_dist=8):
    '''
    Create clustering models, select phoneme features, train classifiers and save everything to a model file so that it can used later for testing.

    Parameters
    ----------

    `models_save_file`: string, file path to which clustering+classifying models are saved;

    `X_formants_train`: array-like, shape = [n_clips, max_frames, n_formant_features]. Formant features for N utterences with fixed number of frames. Pass the actual frames length of each clip as array 'X_frame_lens_train' with the same order.

    `Y_labels_train`: int array, shape = [n_clips]. Labels all clips in X_formants_train in the same order.

    `X_frame_lens_train`: int array, shape = [n_clips]. Array of number of filled frames of each clip out of max_frames (max_frames default=800).
    
    `inst_phoneme_types`: array-like, dtype=int16, shape = [n_models], optional (default=[16, 32, 64]). Cluster numbers for instantaneous (~25ms) phoneme clustering model. Set between 8 to 300 for each cluster model.;
    
    `diff_phoneme_types`:  array-like, dtype=int16, shape = [n_models], optional (default=[16, 32, 64]). Cluster numbers for differential phoneme (~25ms * g_dist) clustering models. Set between 8 to 300 for each cluster model.;

    `K_SD`: float, optional (default=0.0). Feature selection parameters. Set between -1 to 1. It sets the limit of standard deviation below the mean for selecting features within this threshold. Lower value selects more features;

    `g_dist`: unsigned int, optional (default=8). Number for adjacent frames for measuring the change in formant features to calculate differential phoneme features;
    
    Returns
    -------
    `Trained_classifiers`= `ClassifierObject()`, includes multiple trained classifiers, it doesn't need to be passed, function 'Test_model' reads this automatically from `models_save_file`
    '''

    print("Clustering (inst.)", inst_phoneme_types, X_formants_train.shape)
    inst_clust_models = []
    for pt in range(0, len(inst_phoneme_types)):
        #print("Clustering phonemes (inst.) using training clips, Type:", inst_phoneme_types[pt])

        #for clm in range(7):
        pt_clustModel = make_clusters(inst_phoneme_types[pt], X_formants_train, clustering_method=0, print_quality_scores=0, save_plot=False)
        
        #exit()

        inst_clust_models.append(pt_clustModel)


    train_slope_ftrs = formants_slope(X_formants_train, X_frame_lens_train, distance=g_dist)


    print("Clustering (diff.)", diff_phoneme_types, train_slope_ftrs.shape)
    diff_clust_models = []
    for pt in range(0, len(diff_phoneme_types)):
        #print("Clustering phonemes (diff.) using training clips, Type:", diff_phoneme_types[pt])
        slope_clustModel = make_clusters(diff_phoneme_types[pt], train_slope_ftrs, clustering_method=0, print_quality_scores=0, save_plot=False)
        
        diff_clust_models.append(slope_clustModel)

    
    max_clips_train = X_formants_train.shape[0]

    print("Counting phonemes")
    all_ftrs = int(np.sum(inst_phoneme_types) +  np.sum(diff_phoneme_types) + 1) #+ np.sum(comb_phoneme_type**2))
    trainclips_phoneme_count = np.zeros((max_clips_train, all_ftrs,), dtype=np.float)

    #print("Counting phonemes (single). Types:", inst_phoneme_types)
    for pt in range(0, len(inst_phoneme_types)):
        stx = int(np.sum(inst_phoneme_types[0:pt]))
        edx = stx + inst_phoneme_types[pt]
        trainclips_phoneme_count[:, stx:edx] = count_phonemes_in_set(inst_clust_models[pt], X_formants_train, X_frame_lens_train, inst_phoneme_types[pt])


    #print("Counting phonemes (slopes). Types:", diff_phoneme_types)
    for pt in range(0, len(diff_phoneme_types)):
        stx = int(np.sum(inst_phoneme_types) + np.sum(diff_phoneme_types[0:pt]))
        edx = stx + diff_phoneme_types[pt]
        trainclips_phoneme_count[:, stx:edx] = count_phonemes_in_set(diff_clust_models[pt], train_slope_ftrs, X_frame_lens_train, diff_phoneme_types[pt])


    trainclips_phoneme_count[:, all_ftrs - 1] = X_frame_lens_train

    print("Selecting features, K_SD:", K_SD)
    selected_features = select_features(trainclips_phoneme_count, Y_labels_train, K_SD=K_SD, mode=0)

    ClassifierObj = ClassifierObject()
    print("Training Classifiers", "(samples, features):", trainclips_phoneme_count[:, selected_features].shape)
    Trained_classifiers = ClassifierObj.Train(trainclips_phoneme_count[:, selected_features], Y_labels_train)
    

    #Save trained models to pickle file
    import pickle
    tuple_objects = (inst_phoneme_types, diff_phoneme_types, g_dist, inst_clust_models, diff_clust_models, selected_features, Trained_classifiers, np.unique(Y_labels_train))
    pickle.dump(tuple_objects, open(models_save_file, 'wb'))

    #print("Training complete")

    return Trained_classifiers


def Test_model(models_save_file, X_formants_test, Y_labels_test, X_frame_lens_test):

    '''
    Predict emotion classes using array-like features input of n_clips

    Parameters
    ----------

    `models_save_file`: string, file path where clustering+classifying models are saved, use `Train_model()` to create a model file.

    `X_formants_test`: array-like, shape = [n_clips, max_frames, n_formant_features]. Formant features for N utterences with fixed number of frames. Pass the actual frames length of each clip as array 'X_frame_lens_test' with the same order.

    `Y_labels_test`: int array, shape = [n_clips]. Labels all clips in X_formants_test in the same order.

    `X_frame_lens_test`: int array, shape = [n_clips]. Array of number of filled frames of each clip out of max_frames.
    
    Returns
    -------

    `classifiers_results`: Dict list, shape: [{'classifier' : <classifier_name, string>, 'confusion' : <conf_matrix, array-like>, 'UAR' : <uar, float>, 'WAR' : <war, float>}]

    `n_selected_features`: int, number of selected features.
    '''

    #Load clusering and ML models from pickle file
    import pickle
    inst_phoneme_types, diff_phoneme_types, g_dist, inst_clust_models, diff_clust_models, selected_features, Trained_classifiers, trained_unique_labels = pickle.load(open(models_save_file, 'rb'))

    test_slope_ftrs = formants_slope(X_formants_test, X_frame_lens_test, distance=g_dist)
    max_clips_test = X_formants_test.shape[0]

    all_ftrs = int(np.sum(inst_phoneme_types) +  np.sum(diff_phoneme_types) + 1) #+ np.sum(comb_phoneme_type**2))
    testclips_phoneme_count = np.zeros((max_clips_test, all_ftrs,), dtype=np.float)

    #print("Counting phonemes (single). Types:", inst_phoneme_types)
    for pt in range(0, len(inst_phoneme_types)):
        stx = int(np.sum(inst_phoneme_types[0:pt]))
        edx = stx + inst_phoneme_types[pt]
        testclips_phoneme_count[:, stx:edx] = count_phonemes_in_set(inst_clust_models[pt], X_formants_test, X_frame_lens_test, inst_phoneme_types[pt])
    

    #print("Counting phonemes (slopes). Types:", diff_phoneme_types)
    for pt in range(0, len(diff_phoneme_types)):
        stx = int(np.sum(inst_phoneme_types) + np.sum(diff_phoneme_types[0:pt]))
        edx = stx + diff_phoneme_types[pt]
        testclips_phoneme_count[:, stx:edx] = count_phonemes_in_set(diff_clust_models[pt], test_slope_ftrs, X_frame_lens_test, diff_phoneme_types[pt])


    testclips_phoneme_count[:, all_ftrs - 1] = X_frame_lens_test

    classifiers_results = []

    for index, (name, classifier) in enumerate(Trained_classifiers.items()):
        y_pred = Test(classifier, testclips_phoneme_count[:, selected_features])
        conf_matrix = create_conf_matrix(Y_labels_test, y_pred, trained_unique_labels)
        uar, war = get_UAR_WAR(conf_matrix)
        #print(name, uar, war)
        classifiers_results.append({'classifier' : name, 'confusion' : conf_matrix, 'UAR' : uar, 'WAR' : war})
    
    ##################
    #print([chr(x) for x in #label_classes])

    n_selected_features = len(selected_features)
    
    return classifiers_results, n_selected_features


def model_predict_wav_file(models_save_file, test_file):
    '''
    Predict emotion class of a WAV file

    Parameters
    ----------

    `models_save_file`: string, file path where clustering+classifying models are saved, use `Train_model()` to create a model file.

    `test_file`: string, path to the WAV file.

    Returns
    -------

    `classifiers_results`: Dict list of predicted classes by all classifiers, shape = [{'classifier' : <classifier_name>, 'Prediction' : '<label>'}]

    '''

    from FormantsLib.FormantsExtract import Extract_wav_file_formants

    array_frames_by_features, frame_count, signal_length, trimmed_length = Extract_wav_file_formants(test_file, window_length=0.025, window_step=0.01, emphasize_ratio=0.65, f0_min=30, f0_max=4000, max_frames=800, formants=3)

    
    X_formants_test = np.array([array_frames_by_features])
    
    X_frame_lens_test = np.array([frame_count])
    #print("Testing")
    #Load clusering and ML models from pickle file
    import pickle

    inst_phoneme_types, diff_phoneme_types, g_dist, inst_clust_models, diff_clust_models, selected_features, Trained_classifiers, trained_unique_labels = pickle.load(open(models_save_file, 'rb'))


    test_slope_ftrs = formants_slope(X_formants_test, X_frame_lens_test, distance=g_dist)
    max_clips_test = X_formants_test.shape[0]

    all_ftrs = int(np.sum(inst_phoneme_types) +  np.sum(diff_phoneme_types) + 1) #+ np.sum(comb_phoneme_type**2))
    testclips_phoneme_count = np.zeros((max_clips_test, all_ftrs,), dtype=np.float)

    #print("Counting phonemes (single). Types:", inst_phoneme_types)
    for pt in range(0, len(inst_phoneme_types)):
        stx = int(np.sum(inst_phoneme_types[0:pt]))
        edx = stx + inst_phoneme_types[pt]
        testclips_phoneme_count[:, stx:edx] = count_phonemes_in_set(inst_clust_models[pt], X_formants_test, X_frame_lens_test, inst_phoneme_types[pt])
    

    #print("Counting phonemes (slopes). Types:", diff_phoneme_types)
    for pt in range(0, len(diff_phoneme_types)):
        stx = int(np.sum(inst_phoneme_types) + np.sum(diff_phoneme_types[0:pt]))
        edx = stx + diff_phoneme_types[pt]
        testclips_phoneme_count[:, stx:edx] = count_phonemes_in_set(diff_clust_models[pt], test_slope_ftrs, X_frame_lens_test, diff_phoneme_types[pt])


    testclips_phoneme_count[:, all_ftrs - 1] = X_frame_lens_test

    classifiers_results = {}

    for index, (name, classifier) in enumerate(Trained_classifiers.items()):
        y_pred = Test(classifier, testclips_phoneme_count[:, selected_features])
        
        #print(name, [chr(int(x)) for x in y_pred][0])
        classifiers_results[name] = [chr(int(x)) for x in y_pred][0]

    
    return classifiers_results







import Train_Test

def main():

    #Few parameters:
    pt = [16, 32, 64, 128]  #total phoneme clusters, add more than one integer to this list to create multiple models
    K_SD = 0              #float, -1 to +1, feature selection parameters, less K_SD selects more featues
    norm = 0            #Normalization of mel-filter banks

    
    # List of DB names and directory where wav files are stored:
    db_names_paths = [{'DB': "IEMOCAP", 'path' : "C:\\DB\\IEMOCAP_noVideo\\"},]
    # Can add more than one DBs to the list

    '''
    Currently supports these DBs
    {'DB': "EmoDB", 'path' : "C:\\DB\\EMO-DB\\wav\\"},
    {'DB': "RAVDESS", 'path' : "C:\\DB\\RAVDESS_COPY2\\"},
    {'DB': "IEMOCAP", 'path' : "C:\\DB\\IEMOCAP_noVideo\\"},
    {'DB': "ShemoDB", 'path' : "C:\\DB\\shemo\\"},
    {'DB': "DEMoS", 'path' : "C:\\DB\\wav_DEMoS\\DEMOS\\"}
    '''
    
    Train_Test.run_train_and_test(db_names_paths, "results.csv", val_scheme='k-folds', test_size=0.2, folds_n=5, Nm_inst=pt, Nm_diff=pt, K_SD=K_SD, g_dist=6, window_length=0.025, window_step=0.010, emphasize_ratio=0.65, norm=norm, deselect_labels=None)

    exit()

    
    # For cross-corpus validation:
    # Training sets:
    db_names_paths = [{'DB': "RAVDESS", 'path' : "C:\\DB\\RAVDESS_COPY2\\"},]

    #Testing sets
    db_names_paths_2 = [{'DB': "IEMOCAP", 'path' : "C:\\DB\\IEMOCAP_noVideo\\"},]

    
    deselect_labels=['D','F','U','E','R', 'C', 'G', 'B']

    Train_Test.run_train_and_test(db_names_paths, "results.csv", val_scheme='cross-corpus', Nm_inst=pt, Nm_diff=pt, K_SD=K_SD, g_dist=6, window_length=0.025, window_step=0.01, emphasize_ratio=0.65, norm=norm, deselect_labels=deselect_labels, db_names_paths_2=db_names_paths_2)
    

    


    print("Done")
    exit()
    



if __name__ == '__main__':
    main()





"""
-----
Author: Abdul Rehman
License:  The MIT License (MIT)
Copyright (c) 2020, Tabahi Abdul Rehman
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
"""

import numpy as np


class Ix(object):
    '''
    Clip label indices for enumeration - Ignore
    '''
    speaker_id, scenario, sex, emotion, intensity, statement, repetition, frame_count, signal_len, trimmed_len, file_size =0,1,2,3,4,5,6,7,8,9,10
    

def print_database_stats(labels):

    print("Total clips", labels.shape[0])
    print("wav files size (MB)", round(np.sum(labels[:, Ix.file_size])/1000, 2))
    print("Total raw length (min)", round(np.sum(labels[:, Ix.signal_len])/60000, 2))
    print("Total trimmed length (min)", round(np.sum(labels[:, Ix.trimmed_len])/60000, 2))
    print("Avg raw length (s)", round(np.mean(labels[:, Ix.signal_len]/1000), 2))
    print("Avg trimmed length (s)", round(np.mean(labels[:, Ix.trimmed_len]/1000), 2))
    print("Avg. frame count", round(np.mean(labels[:, Ix.frame_count]), 2))
    print("Male Female Clips", np.where(labels[:, Ix.sex]==ord('M'))[0].size, np.where(labels[:, Ix.sex]==ord('F'))[0].size)
    
    unique_speaker_id = np.unique(labels[:, Ix.speaker_id])
    print("Unique speakers: ", len(unique_speaker_id))
    print("Speakers id: ", unique_speaker_id)

    
    unique_classes = np.unique(labels[:, Ix.emotion])
    print("Emotion classes: ", len(unique_classes))
    print("Unique emotions: ", [chr(x) for x in unique_classes])
    
    
    print("Emotion", "N clips", "Total(min)", "Trimmed(min)")
    for this_e in unique_classes:
        select_e = np.where(labels[:, Ix.emotion]==this_e)[0]
        print(chr(this_e), '\t', labels[select_e].shape[0], '\t', round(np.sum(labels[select_e, Ix.signal_len]/1000)/60, 2), '\t', round(np.sum(labels[select_e, Ix.trimmed_len]/1000)/60, 2))

    return len(unique_classes), len(unique_speaker_id)


def save_features_stats(db_name, csv_filename, labels, features):


    import csv
    with open(csv_filename, 'a') as csvFile:
        writer = csv.writer(csvFile, delimiter=',', lineterminator = '\n')
        #writer.writerow(["Emotion", "Combination", "Occurrences"])
        writer.writerow(["DB", "Emotion", "N clips", "f0", "p0", "w0", "d0", "f1", "p1", "w1", "d1", "f2", "p2", "w2", "d2"])
        unique_classes = np.unique(labels[:, Ix.emotion])
        
        print("Mean Values")
        print("Emotion", "freq", "power", "width", "diss")
        for this_e in unique_classes:
            select_e = np.where(labels[:, Ix.emotion]==this_e)[0]

            clips_n = features[select_e].shape[0]
            e_fts = features[select_e]
            this_row = [db_name, str(chr(this_e)), str(clips_n)]
            for i in range(0, 3):
                formant_decay_rate = 0.5**i
                freq = int(np.mean(np.exp(e_fts[:, :, (i*4)][np.where(e_fts[:, :, (i*4)] > 0)] / (200*formant_decay_rate))))
                power = int(np.mean(e_fts[:, (i*4)+1][np.where(e_fts[:, (i*4)+1] > 0)]) / (100*formant_decay_rate) *10)
                width = int(np.mean(e_fts[:, (i*4)+2][np.where(e_fts[:, (i*4)+2] > 0)]) / (5*formant_decay_rate))
                diss = int(np.mean(e_fts[:, (i*4)+3][np.where(e_fts[:, (i*4)+3] > 0)]) / (10*formant_decay_rate))

                this_row.append(str(freq))
                this_row.append(str(power))
                this_row.append(str(width))
                this_row.append(str(diss))
                print(chr(this_e), clips_n, freq, power, width, diss)

            
            writer.writerow(this_row)
            #print(1000, np.log(1000), np.exp(np.log(1000)))

    return



def import_features_from_HDF(storage_file, deselect_labels=None):
    # deselect_labels=['C', 'D', 'F', 'U'])
    print("Reading dataset from file:", storage_file)
    import h5py
    hf = h5py.File(storage_file, 'r')
    lbl = np.array(hf.get('labels'))
    formant_features = np.array(hf.get('features'))

    conditions =  (lbl[:, Ix.scenario]==1) #RAVDESS has 2 scenarios (1=speech, 2=song), select only speech.
    
    if(len(deselect_labels) > 0):
        for em in deselect_labels:
            conditions &= (lbl[:, Ix.emotion]!=ord(em))
    
    selected = np.where(conditions)
    lbl = lbl[selected]
    formant_features = formant_features[selected]

    if(lbl.shape[0]!=formant_features.shape[0]):
        raise Exception("Labels and Features samples size mismatch", lbl.shape[0], formant_features.shape[0])
    
    print ("Clips count:", formant_features.shape[0])

    unique_speaker_id = np.unique(lbl[:, Ix.speaker_id])
    unique_classes = np.unique(lbl[:, Ix.emotion])
    
    return formant_features, lbl, unique_speaker_id, unique_classes



def import_mutiple_HDFs(storage_files, deselect_labels=['C', 'D', 'F', 'U', 'E', 'R', 'G', 'B']):
    # deselect_labels=['C', 'D', 'F', 'U', 'E', 'R', 'G', 'B']
    import os.path as os_path
    import h5py
    print("Reading dataset from file:", storage_files)

    #check if features are already extracted
    if (os_path.isfile(storage_files[0])==False) or (int(os_path.getsize(storage_files[0]))<8000):
        raise Exception ("Formants features for this training set are not extracted yet. Call 'run_train_and_test' for extracting formant features.")
        

    storage_file = storage_files[0]
    hf = h5py.File(storage_file, 'r')
    lbl = np.array(hf.get('labels'))
    formant_features = np.array(hf.get('features'))

    #print(lbl[0:20, Ix.scenario])
    #exit()

    for sn in range(1, len(storage_files)):
        if (os_path.isfile(storage_files[sn])==False) or (int(os_path.getsize(storage_files[sn]))<8000):
            raise Exception ("Formants features for this training set are not extracted yet. Call 'run_train_and_test' for extracting formant features.")
        
        storage_file = storage_files[sn]
        hf = h5py.File(storage_file, 'r')
        lbl = np.concatenate((lbl, np.array(hf.get('labels'))))
        formant_features = np.concatenate((formant_features, np.array(hf.get('features'))))

    select_scence = 1
    if(len(lbl)>1500): select_scence = 2    #don't select scripted from big databases (IEMOCAP, MS-Improv)

    conditions =  ((lbl[:, Ix.scenario]>=select_scence) ) #RAVDESS has 2 scenarios (1=speech, 2=song), select only speech.
    #scenario: 0=unknown, 1=script, 2=improv, 3=radio/TV, 4=elicited, 5=natural, 6=script-in-improv
    
    if(deselect_labels!=None):
        if(len(deselect_labels) > 0):
            for em in deselect_labels:
                conditions &= (lbl[:, Ix.emotion]!=ord(em))
    
    selected = np.where(conditions)
    lbl = lbl[selected]
    formant_features = formant_features[selected]

    if(lbl.shape[0]!=formant_features.shape[0]):
        raise Exception("Labels and Features samples size mismatch", lbl.shape[0], formant_features.shape[0])
    
    print ("Clips count:", formant_features.shape[0])

    unique_speaker_id = np.unique(lbl[:, Ix.speaker_id])
    unique_classes = np.unique(lbl[:, Ix.emotion])

    

    return formant_features, lbl, unique_speaker_id, unique_classes    


    #if(db_name=="IEMOCAP"):
    #    features, labels, u_speakers, u_classes  = HDFread.import_features_from_HDF(features_HDF_file, window_length, window_step, deselect_labels=['D','F','U','E','R'])
        #Deselect some labels from IEMOCAP because these emotions have very few samples.




from setuptools import setup
from setuptools import setuptools

setup(
    name='PhonemeSER',

    version='1.0.3',

    description='Predict speech emotions from wav files.',
    long_description='Predict speech emotions from wav files. Go to Github for more details: https://github.com/tabahi/Phoneme-Converge-SER',
    author="Tabahi - Abdul Rehman",
    author_email="tabahi@hotmail.fr",
    long_description_content_type="text/markdown",
    url="https://github.com/tabahi/Phoneme-Converge-SER",
    packages=setuptools.find_packages(),
    install_requires = ['sklearn', 'pydub', 'wavio'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',

    py_modules=['PhonemeSER', 'examples/predict', 'examples/example', 'Train_test', 'FormantsLib/FormantsExtract', 'FormantsLib/FormatsHDFread']
)




import PhonemeSER
'''
pip install sklearn pydub wavio
'''


def main():
    # Make sure an already trained model file is available.
    model_file = "data/model_EmoD.RAVD.IEMO.Shem.DEMo.MSPI_single-random0.1_5_2_-0.5_16I32I64I12816D32D64D12860.0250.010.65_0.pkl"
    
    test_wav = "263771femaleprotagonist.wav"

    multi_classifiers_results = PhonemeSER.model_predict_wav_file(model_file, test_wav)
    print(multi_classifiers_results)

if __name__ == '__main__':
    main()

